//=============================================================
//
// class			AudioIOStream
// Author:			Grzegorz Lewandowski
// Created:			09/18/97
//
//-------------------------------------------------------------

import java.io.*;

public class AudioIOStream extends InputStream
{
	ByteArrayInputStream in, back;
	ByteArrayOutputStream out;

	boolean empty, eof, eos;

	int bytesRead;

	public AudioIOStream()
	{
		super();
		out = new ByteArrayOutputStream( 8000 );
		in = null;
		back = null;
		empty = true;
		eof = false;
		eos = false;
		bytesRead = 0;
	}

	public synchronized int read() throws IOException
	{
		while( ( in == null ) || ( in.available() == 0 ) )
		{
			if( ( back != null ) && ( back.available() != 0 ) )
			{
				// System.out.println( "AudioIOStream: buffers swapped" );
				in = back;
				back = null;
				empty = true;
				notify();
			}
			else
			{
				if( out.size() == 0 )
				{
					if( eof )
					{
						System.out.println( "AudioIOStream: EOS" );
						eos = true;
						notify();
						return -1;
					}
					else
					{
						// System.out.println( "AudioIOStream: read() waits..." );
						notify();
						try
						{
							wait();
						}
						catch( InterruptedException e )
						{
							System.err.println( "AudioIOStream: InterruptedException " + e.getMessage() );
						}
					}
				}
				else
				{
					emergencyFill();
					break;
				}
			}
		}
		return in.read();
	}
  
	public synchronized int read( byte  b[], int  off, int  len ) throws IOException
	{
		// System.out.println( "AudioIOStream: read " + len + " bytes" );
		while( ( in == null ) || ( in.available() == 0 ) )
		{
			if( ( back != null ) && ( back.available() != 0 ) )
			{
				// System.out.println( "AudioIOStream: buffers swapped" );
				in = back;
				back = null;
				empty = true;
				notify();
			}
			else
			{
				if( out.size() == 0 )
				{
					if( eof )
					{
						System.out.println( "AudioIOStream: EOS" );
						eos = true;
						notify();
						return -1;
					}
					else
					{
						// System.out.println( "AudioIOStream: read(...) waits..." );
						notify();
						try
						{
							wait();
						}
						catch( InterruptedException e )
						{
							System.err.println( "AudioIOStream: InterruptedException " + e.getMessage() );
						}
					}
				}
				else
				{
					// System.out.println( "AudioIOStream: EMERGENCY FILL" );
					emergencyFill();
					break;
				}
			}
		}
		bytesRead += len;
		System.out.println( bytesRead + " bytes read." );
		return in.read( b, off, len );
	}

	private void fill()
	{
		// System.out.println( "AudioIOStream: filling back stream with " + out.size() + " bytes" );
		back = new ByteArrayInputStream( out.toByteArray() );
		empty = false;
		out.reset();
		notify();
	}

	private void emergencyFill()
	{
		// System.out.println( "AudioIOStream: filling in stream with " + out.size() + " bytes");
		in = new ByteArrayInputStream( out.toByteArray() );
		out.reset();
		notify();
	}

	public synchronized void write( byte[] buf, int len )
	{
		while( out.size() > /*1048576*/ /*524288*/ 131072 )
		{
			try
			{
				// System.out.println( "AudioIOStream: write() waits..." );
				wait();
			}
			catch( InterruptedException e )
			{
				System.err.println( "AudioIOStream: Caught InterruptedException: " + e.getMessage() );
			}
			if( empty )
			{
				break;
			}
		}
		out.write( buf, 0, len );
		if( empty )
		{
			fill();
			notify();
		}
	}

	public synchronized void setEOF()
	{
		eof = true;
		notify();
	}

	public synchronized void waitForEOS()
	{
		//System.out.println( "waiting for EOS..." );
		while( !eos )
		{
			try
			{
				wait();
			}
			catch( InterruptedException e )
			{
				System.err.println( "AudioIOStream: Caught InterruptedException: " + e.getMessage() );
			}
		}
	}

	public boolean getEOS()
	{
		//System.out.println( "EOS: " + eos );
		return eos;
	}
}
    
